#ifndef __DERIVATIVE__
#define __DERIVATIVE__

/*!
  \file

\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
 *  This file provides tools to compute x and y derivatives in Fourier
 * space.
 * 
 *  All the functions in this file assume that the arrays are used with
 * a FFT::Support.
*/


namespace FFT{

  //! Compute the x derivative of \p source put it in \p result.
  //! More precisely, \p source is the FFT of a function f and \p result
  //! is the FFT of the x derivative of f.
  //! Can perform in-place computation: just give \p source = \p result.
  template <typename Complex_data_2D_array>
  void x_derivative(const Complex_data_2D_array& source,
		    Complex_data_2D_array* const result);
  
  //! Compute the y derivative of \p source put it in \p result.
  //! More precisely, \p source is the FFT of a function f and \p result
  //! is the FFT of the y derivative of f.
  //! Can perform in-place computation: just give \p source = \p result.
  template <typename Complex_data_2D_array>
  void y_derivative(const Complex_data_2D_array& source,
		    Complex_data_2D_array* const result);



/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/



  

  template <typename Complex_data_2D_array>
  void x_derivative(const Complex_data_2D_array& source,
		    Complex_data_2D_array* const result){

    typedef typename Complex_data_2D_array::value_type complex_type;
    typedef float                                      real_type;
    typedef unsigned int                               size_type;

    const size_type width  = source.x_size();
    const size_type height = source.y_size();
    const size_type half_width  = width/2;

    if ((result->x_size()!=width)||(result->y_size()!=height)){
      result->resize(width,height);
    }

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){

	const real_type X = static_cast<real_type>(x) -
	  width*((x+half_width)/width);
       
	(*result)(x,y) = complex_type(0,2*M_PI*X/width)*source(x,y);
      }
    }

  }


  template <typename Complex_data_2D_array>
  void y_derivative(const Complex_data_2D_array& source,
		    Complex_data_2D_array* const result){

    typedef typename Complex_data_2D_array::value_type complex_type;
    typedef float                                      real_type;
    typedef unsigned int                               size_type;

    const size_type width  = source.x_size();
    const size_type height = source.y_size();
    const size_type half_width  = width/2;

    if ((result->x_size()!=width)||(result->y_size()!=height)){
      result->resize(width,height);
    }

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
      
	const real_type Y = static_cast<real_type>(y);
 
	(*result)(x,y) = complex_type(0,2*M_PI*Y/height)*source(x,y);
      }
    }

  }

}

#endif
