/*! \file

\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim

 *  This file contain code made by Sylvain Paris under supervision of
 * Franois Sillion for his PhD work with <a
 * href="http://www-artis.imag.fr">ARTIS project</a>. ARTIS is a
 * research project in the GRAVIR/IMAG laboratory, a joint unit of
 * CNRS, INPG, INRIA and UJF.
 *
 *  FFT::Support is the base class that performs the FFT transforms.
 * A FFT::Support object has to be created for each domain size. For instance,
 * to handle images 500x500 and 32x128, one has to create a FFT::Support(500,500)
 * and FFT::Support(32,128).
 *
 *  A buffer object is provided to effciently manipulate the temporary variable.
 * For instance, if one has to use several times the Fourier transform of an
 * image, it is more efficient to apply the FFT once and then to use a buffer
 * to store the result in a buffer than to compute the FFT each time. Most of
 * the functions accept either a non-transformed function or a FFT buffer as
 * input parameters.
 *
 *  The array class is assumed to provide the following functions: x_size(),
 * y_size(), resize(unsigned int, unsigned int).
 *
 *  Here is a code example for the convolution f*g=h :
 \code

    // Initialization 
 
    FFT::Support::set_wisdom_file("wisdom.fftw");
    FFT::Support s(f.x_size(),f.y_size());

    // ################
    // ### FFT of f ###
    // ################

    // We need a temporary buffer.
    FFT::Support::buffer_type buffer = support.create_buffer();

    // Store f data in the fftw structures.
    support.load_space_data(f);
    
    // Compute the FFT.
    support.space_to_frequency();
    
    // Copy the result in a safe place.
    support.save_frequency_data_into_buffer(buffer);
    
    
    // ################
    // ### FFT of g ###
    // ################

    // Store g data in the fftw structures.    
    support.load_space_data(g);
    
    // Compute the FFT.
    support.space_to_frequency();



    
    // Multiplication "in place" in the frequency space.    
    support.multiply_frequency_data_by_buffer(buffer,true);


    

    // ###################
    // ### Inverse FFT ###
    // ###################
    
    // Compute inverse FFT transform.
    support.frequency_to_space();

    // Return data into the original structures    
    support.save_space_data(h);


    // Free the buffer memory.
    support.destroy_buffer(buffer);
 \endcode
 * Note that the convolution is already provided in convolution.h.
 */

#ifndef __SUPPORT_FFT__
#define __SUPPORT_FFT__

#include <errno.h>
#include <stdio.h>

#include <sstream>
#include <string>

#include "fftw3.h"

#include "msg_stream.h"


namespace FFT{

  /*
     #################
     # class Support #
     #################
  */


  
  //! Type for support of FFT calculation.
  class Support {
    
  public:

    typedef unsigned int  size_type;

    typedef double        real_type;
    typedef fftw_complex  complex_type;
    
    typedef real_type*    space_data_type;
    typedef complex_type* frequency_data_type;
    typedef void*         buffer_type;
    
    
    //! Needs the \e width and \e height in the space domain of the real
    //! function with which it will deal.
    inline Support(const size_type w,
		   const size_type h);

    inline ~Support();


    //@{
    //! Give the size of the space domain.
    inline size_type x_space_size() const;
    inline size_type y_space_size() const;
    //@}

    //@{
    //! Give the size of the frequency domain.
    inline size_type x_frequency_size() const;
    inline size_type y_frequency_size() const;
    //@}
   
    //! Create a buffer to store data.
    inline buffer_type create_buffer() const;

    //! Destroy the buffer.    
    inline void destroy_buffer(buffer_type buffer) const;



    
    //! Put space data into a 2D array.
    template <typename Real_data_2D_array>
    void save_space_data(Real_data_2D_array* const array) const;

    //! Get space data from a 2D array.
    template <typename Real_data_2D_array>
    void load_space_data(const Real_data_2D_array& array);


    
    
    //! Put frequency data into a 2D array.
    template <typename Complex_data_2D_array>
    void save_frequency_data(Complex_data_2D_array* const array) const;

    //! Get frequency data from a 2D array.
    template <typename Complex_data_2D_array>
    void load_frequency_data(const Complex_data_2D_array& array);


    
    
    //! Put the space data into a buffer.
    inline void save_space_data_into_buffer(buffer_type buffer) const;

    //! Get the space data from a buffer.
    inline void load_space_data_from_buffer(buffer_type buffer);
    



    //! Put the frequency data into a buffer.
    inline void save_frequency_data_into_buffer(buffer_type buffer) const;

    //! Get the frequency data from a buffer.
    inline void load_frequency_data_from_buffer(buffer_type buffer);


    

    //! Compute the FFT.
    inline void space_to_frequency();

    //! Compute the inverse FFT.
    inline void frequency_to_space();

    


    //! Compute a fast, one-by-one, in-place, complex-valued multiplication
    //! of the frequency data.
    inline void multiply_frequency_data_by_buffer(buffer_type buffer,
						  const bool  scale = false);




    

    // ####################
    // # Static functions #
    // ####################


    //! Set the flags for FFTW.
    static inline void set_fftw_flags(const unsigned flags);

    
    //! Set the file name of the wisdom file. "" not to use a file.
    static inline void set_wisdom_file(const std::string file_name);

    
  private:

    //@{
    //! FFTW structure.
    fftw_plan forward, backward;
    //@}

    //! Real data (in space domain).
    space_data_type     space_data;

    //! Complex data (in frequency domain).
    frequency_data_type frequency_data;

    //@{
    //! Size of the real data.
    
    const size_type width;
    const size_type height;
    //@}

    //@{
    //! Size of the complex data.
    
    const size_type support_w;
    const size_type support_h;
    const size_type mem_size;
    //@}
    
    static inline size_type index(const size_type x,
				  const size_type y,
				  const size_type w,
				  const size_type h);
   

    // ####################
    // # Static functions #
    // ####################


    //! See set_fftw_flags()
    static unsigned fftw_flags;

    //! See set_wisdom_file()
    static std::string wisdom_file;

    //! Number of created support.
    static size_type support_number;

    //! Must be called in a support constructor.
    static inline void new_support();

    //! Must be called in a support destructor.
    static inline void delete_support();
  };



/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/


  

  // ##################
  // # Static section #
  // ##################

  // Static values initialization
  
  unsigned           Support::fftw_flags     = FFTW_EXHAUSTIVE;
  std::string        Support::wisdom_file    = std::string("");
  Support::size_type Support::support_number = 0;

  
  

  // ######################
  // # Non-static section #
  // ######################


  void Support::set_fftw_flags(const unsigned flags){
    fftw_flags = flags;
  }

  void Support::set_wisdom_file(const std::string file_name){
    wisdom_file = file_name;
  }



  void Support::new_support(){
    support_number++;

    // Job is done if not the first support or if not file support required.
    if ((support_number > 1)||(wisdom_file.size() == 0)){
      return;
    }

    
    // If it is the first support, try to load the wisdom data.

    
    // Opening the wisdom file.
    FILE* f = fopen(wisdom_file.c_str(),"r");
    
    if (f != NULL){ // System says "ok"
      if(fftw_import_wisdom_from_file(f) == 0){ // But FFTW says "error".
	Message::warning<<"error while reading FFTW wisdom file \""
			<<wisdom_file<<"\""<<Message::done;
      }
      fclose(f);
    }
    else if (errno != ENOENT){ // System says "error".
      Message::warning<<"error while opening (read mode) FFTW wisdom file \""
		      <<wisdom_file<<"\""<<Message::done;
    }
    // Wisdom is now initialized.
  }


  
  void Support::delete_support(){
    support_number--;

    // Job is done if not the last support.
    if ((support_number > 0)||(wisdom_file.size() == 0)){
      return;
    }

    // If it is the last support, try to save the wisdom data.

    // Save wisdom.
    FILE* f = fopen(wisdom_file.c_str(),"w");

    if (f != NULL){ // System says "ok".
      fftw_export_wisdom_to_file(f);
	fclose(f);
    }
    else{ // System says "error".
      Message::warning<<"error while opening (write mode) FFTW wisdom file \""
		      <<wisdom_file<<"\"\n"
		      <<"no file saved"<<Message::done;
    }
  }

  /*!
    If \e wisdom_file is specified, uses the optimized process. This may
    take time at the first run but it goes faster during the next ones.
    Informations are read from and written into \e wisdom_file.

    TODO: if the file does not exist, it segfaults. Use the \e touch shell
    command to workaround.
  */
  Support::Support(const size_type w,
		   const size_type h):
    width(w),
    height(h),
    
    support_w(w),
    support_h(h/2+1),
    mem_size(w*(h/2+1)*sizeof(complex_type)) {

    new_support();
    
    // Pointer to the space data.
    space_data = reinterpret_cast<space_data_type>(fftw_malloc(mem_size));

    // Same pointer to perform "in place" computation.
    frequency_data = reinterpret_cast<frequency_data_type>(space_data);

    
    // We make the best measurement possible because we use the wisdom system.
    forward  = fftw_plan_dft_r2c_2d(width,height,
				    space_data,frequency_data,
				    fftw_flags);
    
    backward = fftw_plan_dft_c2r_2d(width,height,
				    frequency_data,space_data,
				    fftw_flags);
  }





  
  Support::~Support(){

    delete_support();
    
    // Free memory.
    fftw_destroy_plan(forward);
    fftw_destroy_plan(backward);

    fftw_free(space_data);
  }



  
  Support::size_type Support::x_space_size() const{
    return width;
  }


  
  Support::size_type Support::y_space_size() const{
    return height;
  }


  Support::size_type Support::x_frequency_size() const{
    return support_w;
  }


  Support::size_type Support::y_frequency_size() const{
    return support_h;
  }


  
  Support::size_type Support::index(const size_type x,
				    const size_type y,
				    const size_type w,
				    const size_type h){
    return y + x*h;
  }


  
  
  Support::buffer_type Support::create_buffer() const{

    return fftw_malloc(mem_size);
  }

  void Support::destroy_buffer(buffer_type buffer) const{
    
    fftw_free(buffer);
  }

  

  
  /*!
    \e Real_data_2D_array must provide the following functions: resize(),
    and a (x,y) access operator.
  */
  template <typename Real_data_2D_array>
  void Support::save_space_data(Real_data_2D_array* const array) const{

    array->resize(width,height);

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	const size_type i = index(x,y,support_w,2*support_h);

	(*array)(x,y) = space_data[i];
      }
    }
  }

  /*!
    Real_data_2D_array must provide the following functions:
    x_size(), y_size() and a (x,y) access operator.
  */
  template <typename Real_data_2D_array>
  void Support::load_space_data(const Real_data_2D_array& array){

    if ((array.x_size()!=width)||(array.y_size()!=height)){
      Message::error<<"Support::load_space_data : size does not match"
		    <<Message::done;
    }

    for(size_type x=0;x<width;x++){
      for(size_type y=0;y<height;y++){
	const size_type i = index(x,y,support_w,2*support_h);

	space_data[i] = array(x,y);
      }
    }

  }



  
  /*!
    \e Complex_data_2D_array must provide the following functions: resize(),
    and a (x,y) access operator. Complex_data_2D_array::value_type is assumed
    to be compatible with the standard complex<double> structure.
  */
  template <typename Complex_data_2D_array>
  void Support::save_frequency_data(Complex_data_2D_array* const array) const{

    typedef typename Complex_data_2D_array::value_type cplx_type;
    
    array->resize(support_w,support_h);

    for(size_type x=0;x<support_w;x++){
      for(size_type y=0;y<support_h;y++){
	const size_type i = index(x,y,support_w,support_h);

	(*array)(x,y) = cplx_type(frequency_data[i][0],frequency_data[i][1]);
      }
    }
  }

  /*!
    Complex_data_2D_array must provide the following functions:
    x_size(), y_size() and a (x,y) access operator.
    Complex_data_2D_array::value_type is assumed to be compatible
    with the standard complex<double> structure.
  */
  template <typename Complex_data_2D_array>
  void Support::load_frequency_data(const Complex_data_2D_array& array){

    if ((array.x_size()!=support_w)||(array.y_size()!=support_h)){
      Message::error<<"Support::load_frequency_data : size does not match"
		    <<Message::done;
    }

    for(size_type x=0;x<support_w;x++){
      for(size_type y=0;y<support_h;y++){
	const size_type i = index(x,y,support_w,support_h);

	frequency_data[i][0] = array(x,y).real();
	frequency_data[i][1] = array(x,y).imag();
      }
    }

  }



  
  
  void Support::save_space_data_into_buffer(buffer_type buffer) const{
    memcpy(buffer,space_data,mem_size);
  }

  void Support::load_space_data_from_buffer(buffer_type buffer){
    memcpy(space_data,buffer,mem_size);
  }

  

  
  void Support::save_frequency_data_into_buffer(buffer_type buffer) const{
    memcpy(buffer,frequency_data,mem_size);
  }

  void Support::load_frequency_data_from_buffer(buffer_type buffer){
    memcpy(frequency_data,buffer,mem_size);
  }





  void Support::space_to_frequency(){
    fftw_execute(forward);
  }


  void Support::frequency_to_space(){
    fftw_execute(backward);
  }




  void Support::multiply_frequency_data_by_buffer(buffer_type buffer,
						  const bool  scale){
    
    const size_type size = support_w*support_h;
    const frequency_data_type freq_buffer =
      reinterpret_cast<frequency_data_type>(buffer);

    const real_type s = scale ? 1.0/(width*height) : 1.0;
    
    for(size_type i=0;i<size;i++){

      const double a_re = frequency_data[i][0];
      const double a_im = frequency_data[i][1];
      
      const double b_re = freq_buffer[i][0];
      const double b_im = freq_buffer[i][1];
      
      frequency_data[i][0] = (a_re*b_re - a_im*b_im)*s;
      frequency_data[i][1] = (a_re*b_im + a_im*b_re)*s;
      
    }
    
  }

}


#endif
